/* fonts import */
@import url("https://fonts.googleapis.com/css2?family=Nunito+Sans:wght@300;400;700&display=swap");
@import url("https://fonts.googleapis.com/css2?family=Source+Sans+Pro:wght@300;400;600&display=swap");

/* custom themeing */
.wy-body-for-nav {
    --bg-color0: #fff;
    --bg-color1: #fbfafc;
    --layout-color0: #e7e7e7;
    --layout-color1: #9840d2;
    --shadow-0: rgb(0 33 121 / 8%);
    --shadow-1: rgb(0 0 0 / 7%);
    --font-color0: #000;
    --font-color1: #3d3d3d;
    --font-color2: #8e8e8e;
    --font-color3: #cacaca;
    --font-color4: #ffffff;
    --brand-color0: #46096f;
    --brand-color1: #df0982;
    --active-gradient0: linear-gradient(
        45deg,
        var(--brand-color1),
        var(--brand-color0)
    );
    --active-gradient1: linear-gradient(
        rgb(135, 0, 202) -41.11%,
        rgba(216, 164, 250, 0.92) 197.13%
    );
}

.wy-body-for-nav {
    font-family: "Source Sans Pro", sans-serif;
    font-weight: 400;
}

/* Left Navbar */
.wy-nav-side {
    background-color: var(--bg-color0);
    box-shadow: var(--shadow-1) 0px 10px 55px;
}

.wy-side-nav-search {
    background-color: var(--bg-color0);
    color: var(--font-color0);
}

.wy-side-nav-search > a {
    color: var(--font-color0);
    display: flex;
    align-items: flex-start;
    justify-content: center;
    padding-top: 8px;
    margin-left: -14px;
}

.wy-side-nav-search > a::before {
    content: "";
    background-image: url(../logo.png);
    background-size: contain;
    background-repeat: no-repeat;
    width: 28px;
    height: 28px;
    margin-right: 4px;
    position: relative;
    top: 6px;
}

.wy-side-nav-search > div.version,
.wy-side-nav-search > div.switch-menus,
.wy-side-nav-search > div.switch-menus > div.version-switch select {
    color: var(--font-color2) !important;
}

.wy-menu-vertical li.current > a,
.wy-menu-vertical li.on a {
    background-image: var(--active-gradient0);
    background-size: 150%;
    background-position: left;
    color: var(--font-color4);
    transition: background-position 150ms ease;
}

.wy-menu-vertical li.current > a:hover,
.wy-menu-vertical li.on a:hover {
    background-image: var(--active-gradient0);
    background-size: 150%;
    background-position: right;
}

.wy-menu-vertical li.current > a:hover button.toctree-expand,
.wy-menu-vertical li.on a:hover button.toctree-expand {
    color: var(--font-color3);
}

/* left nav bar list item */
.wy-menu-vertical a {
    color: var(--font-color1);
}

.wy-menu-vertical a:hover {
    background-image: var(--active-gradient1);
    color: var(--font-color4);
}

.wy-nav-content {
    background-color: var(--bg-color1);
    max-width: 100dvw;
    max-width: 100vw;
    min-height: 100dvh;
    min-height: 100vh;
}

/* headers */
.rst-content .toctree-wrapper > p.caption,
h1,
h2,
h3,
h4,
h5,
h6,
legend {
    font-family: "Nunito Sans", sans-serif;
    font-weight: 700;
    color: var(--brand-color0);
}

.wy-menu-vertical header,
.wy-menu-vertical p.caption {
    color: var(--brand-color0);
}

.wy-menu-vertical li.current > a button.toctree-expand,
.wy-menu-vertical li.on a button.toctree-expand {
    color: var(--font-color4);
}

/* Style the counter cards */
.card {
    box-shadow: 0 6px 20px 0 var(--shadow-0);
    padding: 16px;
    text-align: center;
    border-radius: 8px;
    background-color: var(--bg-color0);
    transition: transform 150ms ease;
}

.card > h3 {
    color: var(--font-color0);
}

.card:hover {
    transform: scale(1.05);
}

/* buttons */
.btn {
    display: inline-flex;
    gap: 4px;
    -webkit-box-align: center;
    align-items: center;
    -webkit-box-pack: center;
    justify-content: center;
    position: relative;
    box-sizing: border-box;
    -webkit-tap-highlight-color: transparent;
    outline: 0px;
    border: 0px;
    margin: 0px;
    cursor: pointer;
    user-select: none;
    vertical-align: middle;
    appearance: none;
    text-decoration: none;
    font-family: "Source Sans Pro", sans-serif;
    font-weight: 600;
    font-size: 0.875rem;
    line-height: 1.75;
    letter-spacing: 0.02857em;
    min-width: 64px;
    padding: 6px 8px;
    border-radius: 4px;
    text-transform: capitalize;
    color: var(--font-color4);
    background-color: var(--layout-color1);
}

.btn:active {
    padding: 6px 8px;
}

.btn-neutral {
    color: var(--font-color4) !important;
    background-color: var(--layout-color1) !important;
    transition: background-image 150ms linear;
}

.btn-neutral:hover {
    color: var(--font-color4) !important;
    background-image: var(--active-gradient1) !important;
}

.btn-neutral:visited {
    color: var(--font-color4) !important;
}

/* Search box */
.wy-side-nav-search input[type="text"] {
    webkit-transition: 0.6s cubic-bezier(0.165, 0.84, 0.44, 1);
    -o-transition: 0.6s cubic-bezier(0.165, 0.84, 0.44, 1);
    transition: 0.6s cubic-bezier(0.165, 0.84, 0.44, 1);
    -webkit-appearance: none;
    outline: transparent;
    height: 45px;
    padding: 0 30px;
    border-radius: 45px;
    width: 100%;
    outline: none;
    -webkit-box-shadow: none;
    box-shadow: none;
    border: 1px solid var(--layout-color0);
    background-color: var(--bg-color0);
    font-size: 14px;
    line-height: 1.42857143;
    color: var(--font-color1);
    display: inline;
    overflow: visible;
    margin: 0;
    font-family: "Source Sans Pro", sans-serif;
}

/* Foot note */
.rst-content .seealso .admonition-title {
    background: var(--active-gradient0);
}

.rst-content .seealso {
    background: var(--layout-color0);
    border-radius: 8px;
    overflow: hidden;
}

/* common elements */
.rst-content p a {
    -webkit-text-fill-color: transparent;
    text-fill-color: transparent;
    background: var(--active-gradient0);
    -webkit-background-clip: text;
    background-clip: text;
}

.wy-breadcrumbs > li.wy-breadcrumbs-aside > a {
    -webkit-text-fill-color: transparent;
    text-fill-color: transparent;
    background: var(--active-gradient0);
    -webkit-background-clip: text;
    background-clip: text;
}

/* img */
.rst-content img {
    border-radius: 8px;
}

/* highlighted texts */
.rst-content .highlighted {
    -webkit-text-fill-color: var(--font-color4);
    text-fill-color: var(--font-color4);
}

.rst-content code.xref,
.rst-content tt.xref,
a .rst-content code,
a .rst-content tt {
    -webkit-text-fill-color: var(--brand-color0);
    text-fill-color: var(--brand-color0);
}
